# encoding: utf-8
# built-in modules
from __future__ import print_function
import datetime
import logging
import cPickle as pickle
import zlib

# django related setup
if __name__ == "__main__":
    import os
    import sys
    sys.path.append("../")
    os.environ['DJANGO_SETTINGS_MODULE'] = 'ogra_server.settings'

# django imports
from django.utils.translation import ugettext as _
from django.db.models import Max, Min
from django.utils import translation
from django.core.cache import cache

# local imports
from models import DataType, Dimension, DataPoint
from common import DateManipulation, Consts, create_cache_key
from aggregation import Aggregator

class DataRetreivalException(Exception):
    pass


class DataQuery(object):
    """
    This class is responsible for fetching and processing data from the database
    based on the input parameters
    """
    
    TIME_AGGREG_VALUES = (Consts.TIME_AGGREG_NONE,
                          Consts.TIME_AGGREG_LAST,
                          Consts.TIME_AGGREG_AVG,
                          Consts.TIME_AGGREG_MIN,
                          Consts.TIME_AGGREG_MAX,
                          Consts.TIME_AGGREG_DIFF,
                          Consts.TIME_AGGREG_DIFF_REL,
                          Consts.TIME_AGGREG_SUM,
                          )
    

    def __init__(self, data_type, from_date, to_date,
                 fixed_values=None,
                 time_aggregate=Consts.TIME_AGGREG_NONE,
                 series_num_limit=None, time_sampling=None):
        """
        time_aggregate - how to aggregate in time dimension
        time_sampling - if data should be sampled in the time dimension
                        and processed using the aggregation function
        fixed_value - list of length = len(dimensions) with fixed values; 
                      this reduces the effective number of dimensions
        postprocess_bundle - this value is passed to the DataBundle in the
                             postprocess argument 
        """
        self._preprocessed = False
        self.set_params(data_type=data_type)
        self.dimensions = self.data_type.get_dimensions()
        self.set_params(from_date=from_date)
        self.set_params(to_date=to_date)
        self.set_params(time_aggregate=time_aggregate,
                        time_sampling=time_sampling)
        self.set_params(fixed_values=fixed_values)
        
        
    def set_params(self, **kwargs):
        for key, value in kwargs.iteritems():
            # do checks
            if key == 'data_type':
                assert(isinstance(value, DataType))
            elif key == 'from_date':
                assert(value is None or \
                       isinstance(value, datetime.datetime) or \
                       isinstance(value, datetime.date))
            elif key == 'to_date':
                assert(value is None or \
                       isinstance(value, datetime.datetime) or \
                       isinstance(value, datetime.date))
            elif key == 'fixed_values':
                if value is not None:
                    assert(type(value) in (list, tuple))
                    assert(len(value) == len(self.dimensions))
            elif key == "time_aggregate":
                assert(value in self.TIME_AGGREG_VALUES)
            elif key == "time_sampling":
                if value is not None:
                    assert(type(value) in (int, str, unicode))
                    if type(value) in (str, unicode):
                        # it must be properly formatted
                        if not DateManipulation.DATE_RANGE_MATCHER.match(value):
                            raise ValueError("Wrong format of time_sampling.")
            else:
                raise ValueError("Unknown parameter '%s'='%s'" % (key, value))
            setattr(self, key, value)

            
    def preprocess_input(self):
        """
        Takes the input and prepares it for direct usage in queries,
        does value remapping, etc.
        """
        if self.fixed_values:
            fixed_ints = []
            for i, value in enumerate(self.fixed_values):
                if value is not None and type(value) is not int:
                    # int value is presumably already remapped
                    dim = self.dimensions[i]
                    int_val = dim.value_to_int(value)
                else:
                    int_val = value
                fixed_ints.append(int_val)
            self.fixed_values = fixed_ints
        else:
            self.fixed_values = len(self.dimensions) * [None]
        self._preprocessed = True
        
    def _adjust_date(self, date, query_base):
        # get nearest lower value
        try:
            before = query_base.filter(date__lte=date).values("date")\
                                                        .order_by("-date")[0]
        except IndexError:
            before = None
        else:
            before = before['date'].date()
        # get nearest higher value
        try:
            after = query_base.filter(date__gte=date).values("date")\
                                                        .order_by("date")[0]
        except IndexError:
            after = None
        else:
            after = after['date'].date()
        # based on what we got decide what to use
        if not before:
            return after
        elif not after:
            return before
        elif (date-before) < (after-date):
            return before
        else:
            return after 
        
            
    def _fetch_from_db_or_cache(self, queryset):
        cache_key = create_cache_key("db_data", queryset.query)
        ret = cache.get(cache_key)
        if not ret:
            ret = list(queryset)
            # we compress the data because it sometimes goes over the limit
            # of memcached
            cache.set(cache_key, zlib.compress(pickle.dumps(ret, protocol=-1)))
        else:
            ret = pickle.loads(zlib.decompress(ret))
        return ret
    
            
    def fetch_results(self):
        """
        Does the actual querying and data processing
        """
        if not self._preprocessed:
            self.preprocess_input()
        if self.from_date == self.to_date and len(self.dimensions) > 0:
            axes = ["_amount"]
            returned_fields = ["amount"]
        else:
            axes = ["_time","_amount"]
            returned_fields = ["date","amount"]
        limits = {}
        for i, dim in enumerate(self.dimensions):
            attr_name = "value%d" % (i+1)
            fixed_value = self.fixed_values[i]
            if fixed_value is not None:
                # if it is fixed, it belongs to limits, not output
                limits[attr_name] = fixed_value
            else:
                # this value will be part of the output
                axes.append(dim)
                returned_fields.append(attr_name)
        # time aggregation
        if self.time_sampling is None and \
           self.time_aggregate in (Consts.TIME_AGGREG_DIFF,
                                   Consts.TIME_AGGREG_DIFF_REL):
            # base for the query
            query = self.data_type.datapoint_set.order_by("date")
            if not self.from_date or not self.to_date:
                # we need to obtain the from data from the database
                dates = DataPoint.objects.filter(data_type=self.data_type)\
                        .aggregate(min_date=Min('date'), max_date=Max('date'))
            if not self.from_date:
                if dates['min_date']:
                    self.from_date = dates['min_date'].date()
                else:
                    raise DataRetreivalException("No data for data_type")
            if not self.to_date:
                if dates['max_date']:
                    self.to_date = dates['max_date'].date()
                else:
                    raise DataRetreivalException("No data for data_type")
            # we query both dates separately and attempt to fix the date if
            # it is not available
            query_base = query.filter(**limits)
            query = query_base.values_list(*returned_fields)
            start_data = list(query.filter(date=self.from_date))
            if not start_data:
                self.from_date = self._adjust_date(self.from_date, query_base)
                start_data = list(query.filter(date=self.from_date))
            end_data = list(query.filter(date=self.to_date))
            if not end_data:
                self.to_date = self._adjust_date(self.to_date, query_base)
                end_data = list(query.filter(date=self.to_date))
            result = start_data + end_data
        else:
            # base for the query
            if self.time_sampling in ("1m", "1y"):
                query = self.data_type.datapointaggreg_set\
                                    .filter(time_period="1m").order_by("date")
            else:
                query = self.data_type.datapoint_set.order_by("date")
            if self.from_date:
                limits['date__gte'] = self.from_date
            if self.to_date:
                limits['date__lte'] = self.to_date
            query = query.filter(**limits)
            list_query = query.values_list(*returned_fields)
            result = self._fetch_from_db_or_cache(list_query)
        # return the result as a DataBundle instance
        kwargs = dict(amount_dim=self.data_type.amount_dim,
                      data_time_aggreg=self.time_aggregate,
                      time_sampling=self.time_sampling)
        if len(axes) == 2:
            bundle = DataBundle2d(axes, result, **kwargs)
            if self.time_sampling is None and \
               self.time_aggregate in (Consts.TIME_AGGREG_DIFF,
                                       Consts.TIME_AGGREG_DIFF_REL):
                bundle = bundle.transform(self.time_aggregate)
        elif len(axes) == 3:
            bundle = DataBundle3d(axes, result, **kwargs)
            if self.time_sampling is None and \
               self.time_aggregate in (Consts.TIME_AGGREG_DIFF,
                                       Consts.TIME_AGGREG_DIFF_REL):
                bundle = bundle.transform(self.time_aggregate)
        else:
            raise NotImplementedError("Not yet implemented")
        return bundle
    
    
class DataBundle(object):
    
    def __init__(self, dimensions, data, amount_dim=None,
                 data_time_aggreg=Consts.TIME_AGGREG_NONE, time_sampling=None):
        self.raw_data = data
        self.dimensions = dimensions
        self.amount_dim = amount_dim
        self.data_time_aggreg = data_time_aggreg
        self.time_sampling = time_sampling
        self.data = []
        self.series_dim = None
        self.key_dim = None
        
        
    def __str__(self):
        return unicode(self).encode('utf-8')
        
    def __unicode__(self):
        dim_num = len(self.dimensions)
        rec_num = len(self.data)
        dims = u",".join([unicode(x) for x in self.dimensions])
        return u"{0:d} {1:d}D records; dimensions: {2}".format(rec_num, dim_num,
                                                               dims)
        
    def _remap_values(self):
        if isinstance(self.series_dim, Dimension) and \
           self.series_dim.type == Dimension.TYPE_TEXT:
            series_remap = self.series_dim.get_int_to_str_remap(
                                                lang=translation.get_language())
            series_remap = self.series_dim.\
                get_int_to_str_and_custom_values_remap(
                                                lang=translation.get_language())
            series_remap["other"] = (_("Other"), "", 0)
        else:
            series_remap = None
        # main key remap
        if isinstance(self.key_dim, Dimension) and \
           self.key_dim.type == Dimension.TYPE_TEXT:
            key_remap = self.key_dim.get_int_to_str_remap(
                                                lang=translation.get_language())
            key_remap["other"] = _("Other")
        else:
            key_remap = None
        # do the remap
        if series_remap or key_remap:
            new_data = []
            for main_key, record in self.data:
                new_record = {}
                if key_remap:
                    new_main_key = key_remap.get(main_key, str(main_key))
                else:
                    new_main_key = main_key
                if series_remap:
                    for key, value in record.iteritems():
                        new_record[series_remap.get(key, str(key))] = value
                else:
                    new_record = record
                new_data.append((new_main_key, new_record))
            self.data = new_data
            
    @classmethod
    def group_data(cls, records, grouping_idx, value_idx):
        """
        Data must be sorted by grouping_idx
        """
        if not records:
            raise StopIteration
        series_idxs = []
        if len(records[0]) > 2:
            for i in range(len(records[0])):
                if i != grouping_idx and i != value_idx:
                    series_idxs.append(i)
        last_grouping_value = None
        row = {}
        for record in records:
            grouping_value = record[grouping_idx]
            value_value = record[value_idx]
            if len(series_idxs) == 0:
                series_value = None
            elif len(series_idxs) == 1:
                series_value = record[series_idxs[0]]
            else:
                series_value = tuple([record[idx] for idx in series_idxs])
            if not last_grouping_value:
                last_grouping_value = grouping_value
            if grouping_value != last_grouping_value:
                yield (last_grouping_value, row)
                last_grouping_value = grouping_value
                row = {}
            row[series_value] = value_value
        yield (last_grouping_value, row)
        
        
    def postprocess(self, limit_series=0, series_set=None, create_other=None,
                    scale=Consts.SCALE_NORMAL):
        """
        This method references some internal methods that have to be 
        overridden in a child
        """
        self._postprocess_sampling(sampling=self.time_sampling)
        self._postprocess_scale(scale=scale)
        self._postprocess_limit_series(limit_series=limit_series,
                                       series_set=series_set,
                                       create_other=create_other)
        self._postprocess_add_zeros()
        self._remap_values()
        

    def _postprocess_limit_series(self, limit_series=0, series_set=None,
                                  create_other=None):
        pass
    
    
    def _postprocess_scale(self, scale=Consts.SCALE_NORMAL):
        pass


    def _postprocess_sampling(self, sampling=None):
        if self.key_dim != "_time" or sampling is None:
            # no sampling for data not depending on time
            # or if there is not sampling method
            return
        grouped = {}
        last_main_key = None
        group_breaks = DateManipulation.get_time_breaker_fn(sampling)
        aggreg_fn = Aggregator.get_aggregate_fn(self.data_time_aggreg)
        new_data = []
        for main_key, record in self.data:
            if last_main_key is None or group_breaks(last_main_key, main_key):
                new_record = {}
                for key, values in grouped.iteritems():
                    new_value = aggreg_fn(values)
                    new_record[key] = new_value
                if new_record:
                    new_data.append((last_main_key, new_record))
                grouped = {}
            for key, value in record.iteritems():
                if key not in grouped:
                    grouped[key] = []
                grouped[key].append(value)
            last_main_key = main_key
        if grouped:
            new_record = {}
            for key, values in grouped.iteritems():
                new_value = aggreg_fn(values)
                new_record[key] = new_value
            if new_record:
                new_data.append((last_main_key, new_record))
        self.data = new_data

        
    def _postprocess_add_zeros(self):
        pass

    
    def gen_data_table(self):
        """
        Generator for individual rows of a data table, should be overriden
        in children
        """
        return None
    
    def get_stored_time_range(self):
        if self.key_dim != "_time":
            return None, None
        if not self.data:
            return None, None
        else:
            return self.data[0][0], self.data[-1][0]

    
    def get_series_description(self):
        """
        Should return a list of series in form (name, color, order) and a
        dictionary with a DataTable compatible description of the series
        Should be implemented in a child.
        """
        return None, None
    
    def transform(self, transformation):
        """
        Performs a transformation on the data and returns a new DataBundle
        with the transformed data
        """
        return None
    
    def _get_time_dim_description(self):
        """
        A common place for children to get description of a time dimension
        based on sampling method
        """
        if not self.time_sampling:
            return ("date", _('Time'))
        elif self.time_sampling.endswith("m"):
            return ("string", _('Month'))
        elif self.time_sampling.endswith("y"):
            return ("string", _('Year'))
        else:
            raise ValueError("Unknown sampling '%s'" % self.time_sampling)
        
    def _format_main_key(self, key):
        if isinstance(key, datetime.date) or isinstance(key, datetime.datetime):
            if self.time_sampling:
                return DateManipulation.str_date(key, 
                                            rounding=self.time_sampling[-1])
        return key


class DataBundle1d(DataBundle):
    """
    DataBundle specialized for 1D data
    """
    
    def __init__(self, dimensions, data, amount_dim=None,
                 data_time_aggreg=Consts.TIME_AGGREG_NONE,
                 time_sampling=None):
        super(DataBundle1d, self).__init__(dimensions, data,
                                           amount_dim=amount_dim,
                                           data_time_aggreg=data_time_aggreg,
                                           time_sampling=time_sampling)
        if len(self.dimensions) != 1:
            raise Exception("This class can handle only 1D data, use a "
                            "different DataBundle subclass for %dD data" %\
                            len(self.dimensions))
        if not "_amount" in self.dimensions:
            raise Exception("There must be an amount dimension "
                            "between dimensions")
        self.idx_amount = 0
        self.data = dict(data)
        self.series_dim = None
    
    
    def gen_data_table(self):
        amount = self.data[None]
        yield {"_amount": amount, "_main_key": ""}

    
    def get_series_description(self):
        description = {}
        if isinstance(self.amount_dim, Dimension):
            amount_name = self.amount_dim.name
        else:
            amount_name = self.amount_dim or _("Amount")
        description["_amount"] = ('number', amount_name)
        description["_main_key"] = ('string', "")
        return None, description
    

class DataBundle2d(DataBundle):
    """
    DataBundle specialized for 2D data
    """
    
    def __init__(self, dimensions, data, amount_dim=None,
                 data_time_aggreg=Consts.TIME_AGGREG_NONE,
                 time_sampling=None):
        super(DataBundle2d, self).__init__(dimensions, data,
                                           amount_dim=amount_dim,
                                           data_time_aggreg=data_time_aggreg,
                                           time_sampling=time_sampling)
        if len(self.dimensions) != 2:
            raise Exception("This class can handle only 2D data, use a "
                            "different DataBundle subclass for %dD data" %\
                            len(self.dimensions))
        if not "_amount" in self.dimensions:
            raise Exception("There must be an amount dimension "
                            "between dimensions")
        self.idx_amount = self.dimensions.index("_amount")
        self.idx_var = 1 if self.idx_amount == 0 else 0
        self.data = list(self.group_data(self.raw_data, self.idx_var,
                                         self.idx_amount))
        self.series_dim = None
        self.key_dim = self.dimensions[self.idx_var]
    
    
    def _postprocess_limit_series(self, limit_series=0, series_set=None,
                                  create_other=None):
        # time dependent data is never limited in series
        if self.key_dim != "_time" and (series_set or limit_series > 0):
            # prepare some parameters
            if create_other is False or \
               self.data_time_aggreg == Consts.TIME_AGGREG_DIFF_REL:
                # in case of relative difference we do not compute other
                # as it would not make sense
                series_num = limit_series
                do_other = False
            else:
                series_num = limit_series-1
                do_other = True
            # check the data
            series_sums = {}
            for main_key, record in self.data:
                value = record[None]
                series_sums[main_key] = series_sums.get(main_key, 0) + value
            all_series = set(series_sums.keys())
            if not series_set and len(all_series) > limit_series:
                # no preselection, but more series than limit
                series_set = set(series_sums.keys())
            if series_set and \
               (limit_series == 0 or len(series_set) > len(all_series)):
                do_other = False
                series_num = limit_series
            # else the series_set remains None
            # if we do not have a series_set, we can short-circuit, otherwise
            # we do the work
            if series_set:
                if limit_series > 0 and len(series_set) > limit_series:
                    # do not bother if we are in the limit
                    to_sort = [(series_sums[series], series)
                               for series in series_set]
                    to_sort.sort(reverse=True)
                    allowed_series = set([series for val_sum,series
                                          in to_sort[:series_num]])
                else:
                    allowed_series = series_set
                # the actual data filtering
                new_data = []
                other = 0
                for main_key, record in self.data:
                    if main_key in allowed_series:
                        new_data.append((main_key, record))
                    else:
                        other += record[None]
                if do_other:
                    new_data.append(("other", {None: other}))
                self.data = new_data
                

    def _postprocess_scale(self, scale=Consts.SCALE_NORMAL):
        if scale in (Consts.SCALE_RELATIVE, Consts.SCALE_RELATIVE_100):
            # relative scaling is the same in both cases
            if scale == Consts.SCALE_RELATIVE_100:
                factor = 100.0
                round_to = 4
            else:
                factor = 1.0
                round_to = 6
            if self.key_dim == "_time":
                # this would not work with time dependent data
                new_data = []
                for main_key, record in self.data:
                    new_data.append((main_key, {None: factor}))
                self.data = new_data
            else:
                total = 0
                for main_key, record in self.data:
                    total += record[None]
                if total > 0:
                    new_data = []
                    for main_key, record in self.data:
                        value = round(factor*record[None]/total, round_to)
                        new_data.append((main_key, {None: value}))
                    self.data = new_data


    def gen_data_table(self):
        for main_key, rec in self.data:
            amount = rec[None]
            yield {"_main_key": self._format_main_key(main_key),
                   "_amount": amount}

    
    def get_series_description(self):
        description = {}
        if isinstance(self.amount_dim, Dimension):
            amount_name = self.amount_dim.name
        else:
            amount_name = self.amount_dim or _("Amount")
        description["_amount"] = ('number', amount_name)
        var_dim = self.dimensions[self.idx_var]
        if var_dim == '_time':
            description["_main_key"] = self._get_time_dim_description()
        else:
            description["_main_key"] = (var_dim.get_type_string(), var_dim.name)
        return None, description
    
    def transform(self, transformation):
        """
        Performs a transformation on the data and returns a new DataBundle
        with the transformed data
        """
        if transformation in (Consts.TIME_AGGREG_DIFF,
                              Consts.TIME_AGGREG_DIFF_REL):
            if len(self.data) < 2:
                raise DataRetreivalException(
                    "Cannot do difference with %d intervals" % len(self.data))
            elif len(self.data) > 2:
                logging.warn("Ineffective difference with %d time points" %\
                             len(self.data))
            start = self.data[0][1]
            end = self.data[-1][1]
            result = []
            for key, value1 in start.iteritems():
                value2 = end.get(key)
                if value2 is not None:
                    if transformation == Consts.TIME_AGGREG_DIFF:
                        result.append((key, value2-value1))
                    elif transformation == Consts.TIME_AGGREG_DIFF_REL:
                        if value1 != 0:
                            # ignore values starting with 0 at day 1
                            result.append((key,
                                           round(100.0*(value2-value1)/value1, 4)))
            if transformation == Consts.TIME_AGGREG_DIFF:
                amount_dim = self.amount_dim
            elif transformation == Consts.TIME_AGGREG_DIFF_REL:
                amount_dim = self.amount_dim
                #amount_dim = _("Percent difference")
            return DataBundle1d([self.dimensions[self.idx_amount]],
                                 result,
                                 amount_dim=amount_dim,
                                 data_time_aggreg=transformation)
    

class DataBundle3d(DataBundle):
    """
    DataBundle specialized for 3D data
    """
    
    def __init__(self, dimensions, data, amount_dim=None,
                 data_time_aggreg=Consts.TIME_AGGREG_NONE,
                 time_sampling=None):
        super(DataBundle3d, self).__init__(dimensions, data,
                                           amount_dim=amount_dim,
                                           data_time_aggreg=data_time_aggreg,
                                           time_sampling=time_sampling)
        if len(self.dimensions) != 3:
            raise Exception("This class can handle only 3D data, use a "
                            "different DataBundle subclass for %dD data" %\
                            len(self.dimensions))
        self.idx_series = None
        self.idx_time = None
        self.idx_amount = None
        self.series_dim = None
        self.key_dim = None
        for i, dim in enumerate(self.dimensions):
            if isinstance(dim, Dimension):
                self.idx_series = i
                self.series_dim = dim
            elif dim == "_time":
                self.idx_time = i
                self.key_dim = "_time"
            elif dim == "_amount":
                self.idx_amount = i
            else:
                raise Exception("Unknown dimension '%s'" % dim)
        if self.idx_series is None:
            raise Exception("There must be at one Dimension instance "
                            "between dimensions")
        if self.idx_time is None:
            raise Exception("There must be a time dimension "
                            "between dimensions")
        if self.idx_amount is None:
            raise Exception("There must be an amount dimension "
                            "between dimensions")
        self.data = list(self.group_data(self.raw_data, self.idx_time,
                                         self.idx_amount))
        

    def __unicode__(self):
        dim_num = len(self.dimensions)
        rec_num = len(self.data)
        dims = u",".join([unicode(x) for x in self.dimensions])
        return u"{0:d} {1:d}D records; dimensions: {2}".format(rec_num, dim_num,
                                                               dims)


    def _postprocess_limit_series(self, limit_series=0, series_set=None,
                                  create_other=None):
        """
        Does value remapping, etc.
        limit_series - 0 means all series, anything else means X series with
        series_set - defines a set of allowed series values, rest is discarded
        """
        if series_set or limit_series > 0:
            if series_set:
                series_set = set(series_set)
            do_other = True
            # lets find all the cases where we do not want to count 'other'
            if create_other is False or \
               self.data_time_aggreg == Consts.TIME_AGGREG_DIFF_REL:
                # in case of relative difference we do not compute other
                # as it would not make sense
                do_other = False
            # prepare some parameters
            if do_other:
                series_num = limit_series-1
            else:
                series_num = limit_series
            # check the data
            series_sums = {}
            for main_key, record in self.data:
                for key, value in record.iteritems():
                    series_sums[key] = series_sums.get(key, 0) + value
            usable_series = set(series_sums.keys())
            if series_set:
                # we discard things in series set which are not in usable_series 
                series_set &= usable_series
                usable_series = series_set
            if not usable_series:
                # no usable series - probably no data for the current selection
                self.data = []
                return
            if not series_set and len(usable_series) > limit_series:
                # no preselection, but more series than limit
                series_set = usable_series
            # else the series_set remains None
            # if we do not have a series_set, we can short-circuit, otherwise
            # we do the work
            if series_set:
                if limit_series > 0 and len(series_set) > limit_series:
                    # do not bother if we are in the limit
                    to_sort = [(series_sums[series], series)
                               for series in series_set]
                    to_sort.sort(reverse=True)
                    allowed_series = set([series for _val_sum,series
                                          in to_sort[:series_num]])
                else:
                    allowed_series = series_set
                    do_other = False
                # the actual data filtering
                new_data = []
                for main_key, record in self.data:
                    new_record = {}
                    other = 0
                    for key, value in record.iteritems():
                        if key in allowed_series:
                            new_record[key] = value
                        elif key in usable_series:
                            other += value
                        else:
                            # we discard this value
                            pass
                    if do_other:
                        new_record["other"] = other
                    new_data.append((main_key, new_record))
                self.data = new_data


    def _postprocess_scale(self, scale=Consts.SCALE_NORMAL):
        if scale in (Consts.SCALE_RELATIVE, Consts.SCALE_RELATIVE_100):
            # relative scaling is the same in both cases
            if scale == Consts.SCALE_RELATIVE_100:
                factor = 100.0
                round_to = 4
            else:
                factor = 1.0
                round_to = 6
            new_data = []
            for main_key, record in self.data:
                total = sum(record.values(), 0)
                if total == 0:
                    new_record = record
                else:
                    new_record = {}
                    for key, value in record.iteritems():
                        new_record[key] = round(factor*value/total, round_to)
                new_data.append((main_key, new_record))
            self.data = new_data

    
    def _postprocess_add_zeros(self):
        series = set()
        for _main_key, record in self.data:
            series |= set(record.keys())
        for _main_key, record in self.data:
            for series_id in series:
                if series_id not in record:
                    record[series_id] = 0

    
    def gen_data_table(self):
        for main_key, record in self.data:
            new_record = {}
            for key, value in record.iteritems():
                new_record[key[0]] = value
            new_record['_main_key'] = self._format_main_key(main_key)
            yield new_record

    
    def get_series_description(self):
        description = {}
        series_names = set()
        series = set()
        for main_key, record in self.data:
            series_names |= set([key[0] for key in record.keys()])
            series |= set(record.keys())
        for name in series_names:
            description[name] = ('number', name)
        if "_time" in self.dimensions:
            description["_main_key"] = self._get_time_dim_description()
        else:
            description["_main_key"] = ("string", _('Who knows'))
        return series, description

    
    def transform(self, transformation):
        """
        Performs a transformation on the data and returns a new DataBundle
        with the transformed data
        """
        if transformation in (Consts.TIME_AGGREG_DIFF,
                              Consts.TIME_AGGREG_DIFF_REL):
            if len(self.data) < 2:
                raise DataRetreivalException(
                    "Cannot do difference with %d intervals" % len(self.data))
            elif len(self.data) > 2:
                logging.warn("Ineffective difference with %d time points" %\
                             len(self.data))
            start = self.data[0][1]
            end = self.data[-1][1]
            result = []
            all_keys = set(start.keys()) | set(end.keys())
            for key in all_keys:
                value1 = start.get(key, 0)
                value2 = end.get(key, 0)
                if transformation == Consts.TIME_AGGREG_DIFF:
                    result.append((key, value2-value1))
                elif transformation == Consts.TIME_AGGREG_DIFF_REL:
                    if value1 != 0:
                        # ignore values starting with 0 at day 1
                        result.append((key,
                                       round(100.0*(value2-value1)/value1, 4)))
            if transformation == Consts.TIME_AGGREG_DIFF:
                amount_dim = self.amount_dim
            elif transformation == Consts.TIME_AGGREG_DIFF_REL:
                amount_dim = self.amount_dim
                #amount_dim = _("Percent difference")
            return DataBundle2d([self.dimensions[self.idx_series],
                                 self.dimensions[self.idx_amount]],
                                 result,
                                 amount_dim=amount_dim,
                                 data_time_aggreg=transformation)
                
        
    
if __name__ == "__main__":
    from pprint import pprint
    import time
    
    data_type = DataType.objects.get(code_name="hosting")
    from_time = datetime.date(2011, 9, 1)
    to_time = datetime.date(2012, 10, 10)
    #from_time = to_time
    drq = DataQuery(data_type, from_time, to_time,
                    time_aggregate=Consts.TIME_AGGREG_DIFF_REL,
                    time_sampling="1m")
    #drq.set_params(fixed_values=[u'Počet domén'])
    #drq.set_params(fixed_values=['mx',None])
    drq.set_params(fixed_values=['mx',"ACTIVE 24"])
    #drq.set_params(fixed_values=['mx',"ACTIVE 24"])
    t = time.time()
    bundle = drq.fetch_results()
    print("Retreive: {0:.1f}ms".format(1000*(time.time() - t)))
    print(bundle)
    t = time.time()
    bundle.postprocess()
    #pprint(bundle.raw_data)
    #t = time.time()
    #bundle.postprocess(series_set=set([92]))
    #print(bundle)
    #print("Postprocess: {0:.1f}ms".format(1000*(time.time() - t)))
    pprint(bundle.get_series_description())
    for rec in bundle.gen_data_table():
        print(rec)
    