# encoding: utf-8

# built-in modules
import datetime
import logging
import locale
locale.setlocale(locale.LC_ALL, '')

# Django imports
from django.http import Http404
from django.utils.translation import ugettext_lazy as _, ugettext
from django.db.models import Max, Min
from django.utils import translation

# third party modules
import gviz_api

# local imports
from stats import models
from data_mining import DataQuery, DataRetreivalException
from common import DateManipulation, Consts, DateInterval, querydict_to_dict

logger = logging.getLogger("stats.views")

class DataRequest(object):
    
    KEY_SLICING_DIM = "sd"
    KEY_SLICING_VALUE = "sv"
    KEY_REF_DATE = "rd"
    KEY_DATE_RANGE = "dr"
    KEY_TIME_PROCESSING = "tp"
    KEY_SERIES_SET = "ss"
    KEY_DATA_SCALING = "ds" 
    KEY_DISPLAY_AS = "da"
    KEY_SERIES_LIMIT = "sl"
    KEY_INCLUDE_OTHER = "io"
    
    # date range definitions
    DEFAULT_DATE_RANGE = "1y"
    DATE_RANGES = (("0d", _("selected day")),
                   ("1m", _("one month")),
                   ("3m", _("three months")),
                   ("6m", _("six months")),
                   ("1y", _("one year")),
                   ("3y", _("three years")),
                   ("all", _("maximum")),
                   )
    VALID_DATE_RANGE_KEYS = [v[0] for v in DATE_RANGES]
    
    # time processing definitions
    DEFAULT_TIME_PROCESSING = "i-1m"
    TIME_PROCESSINGS = (("i-1d", _("individual days")),
                        ("i-1m", _("individual months")),
                        ("i-1y", _("individual years")),
                        ("diff", _("difference")),
                        ("reldiff", _("relative difference")),
                        )
    VALID_TIME_PROCESSING_KEYS = [v[0] for v in TIME_PROCESSINGS]
    
    # display as
    DEFAULT_DISPLAY_AS = "chart"
    DISPLAY_AS_OPTS = (("chart", _("chart")),
                       ("table", _("table")),
                       )
    VALID_DISPLAY_AS_KEYS = [v[0] for v in DISPLAY_AS_OPTS]
    
    # scaling methods
    VALID_SCALING_KEYS = ["normal","relative","log"]
    DEFAULT_SCALING = "normal"
    
    DIM_KEY_REMAP = {"_time": _("Time")}
    
    SERIES_LIMIT = 15
    
    class IncompatibleSettingsError(Exception):
        pass
    
    class ResultPackage:
        pass

    def __init__(self, params, statistics):
        if isinstance(statistics, models.Statistics):
            self.statistics = statistics
        elif type(statistics) in (str, unicode):
            try:
                self.statistics = models.Statistics.objects\
                                                .select_related("data_type")\
                                                .get(code_name=statistics)
            except models.DataType.DoesNotExist:
                raise Http404
        else:
            raise ValueError("statistics arg must be either Statistics "
                             "instance or code name")
        if self.statistics.data_type is None:
            # this is offline statistics, cannot process it
            raise Http404
        
        self.data_type = self.statistics.data_type
        # check if there is some data at all
        if not self.data_type.has_data():
            raise Http404
        self.dimensions = self.data_type.get_dimensions()

        # merge default_params with params
        if self.statistics.default_params:
            def_params = self.statistics.default_params.get_data_request_params()
        else:
            def_params = {}
        new_params = querydict_to_dict(params)
        def_params.update(new_params)
        params = def_params

        # read parameters
        req_slicing_dim_name = params.get(self.KEY_SLICING_DIM)
        req_slicing_value = params.get(self.KEY_SLICING_VALUE)
        req_ref_date = params.get(self.KEY_REF_DATE)
        req_date_range = params.get(self.KEY_DATE_RANGE,
                                    self.DEFAULT_DATE_RANGE)
        req_time_processing = params.get(self.KEY_TIME_PROCESSING,
                                         self.DEFAULT_TIME_PROCESSING)
        req_series_set = params.get(self.KEY_SERIES_SET, 0)
        if type(req_series_set) is not list:
            req_series_set = [req_series_set]
        req_data_scaling = params.get(self.KEY_DATA_SCALING,
                                      self.DEFAULT_SCALING)
        req_display_as = params.get(self.KEY_DISPLAY_AS,
                                    self.DEFAULT_DISPLAY_AS)
        req_series_limit = params.get(self.KEY_SERIES_LIMIT)
        req_include_other = params.get(self.KEY_INCLUDE_OTHER)
        # sanitize the parameters
        if type(req_include_other) is bool:
            self.include_other = req_include_other
        elif type(req_include_other) is int:
            self.include_other = bool(req_include_other)
        elif req_include_other in ("yes", "true"):
            self.include_other = True
        elif req_include_other in ("no", "false"):
            self.include_other = False
        else:
            self.include_other = None
        # series limit
        if req_series_limit:
            try:
                self.series_limit = int(req_series_limit)
            except ValueError:
                self.series_limit = self.SERIES_LIMIT
        else:
            self.series_limit = self.SERIES_LIMIT
        if req_slicing_value:
            try:
                req_slicing_value = int(req_slicing_value)
            except ValueError:
                pass
        # choose selected dim
        if len(self.dimensions) > 1:
            # we have multi dimensional data, we need to slice it somehow
            if type(req_slicing_dim_name) is int and \
               len(self.dimensions) > req_slicing_dim_name:
                slicing_dim = self.dimensions[req_slicing_dim_name]
            elif req_slicing_dim_name:
                for dim in self.dimensions:
                    if dim.code_name == req_slicing_dim_name:
                        slicing_dim = dim
                        break
                else:
                    slicing_dim = self.dimensions[0]
            else:
                slicing_dim = self.dimensions[0]
            slicing_dim_values = slicing_dim.get_int_to_str_remap(
                                                lang=translation.get_language())
        else:
            slicing_dim = None
            slicing_dim_values = {}
        # choose the selected req_slicing_value
        slicing_value = None
        if slicing_dim_values:
            if req_slicing_value:
                if type(req_slicing_value) is int:
                    if req_slicing_value in slicing_dim_values:
                        slicing_value = req_slicing_value
                else:
                    # it is a string, we need to try to map it back to int
                    slicing_value = slicing_dim.value_to_int(req_slicing_value)
            # we did not get any slicing value, but we need one
            if slicing_value is None:
                slicing_value = sorted(slicing_dim_values.keys())[0]
        # possible series
        possible_series = {}
        series_dim = None
        if self.dimensions:
            for dim in self.dimensions:
                if dim != slicing_dim:
                    series_dim = dim
                    break
            if series_dim:
                possible_series = series_dim.get_int_to_str_remap(
                                                lang=translation.get_language())
        # series_set
        series_set = None
        if req_series_set:
            if (0 in req_series_set) or '0' in req_series_set:
                # 0 means top X or all (if there is less than X)
                series_set = None
            elif (-1 in req_series_set) or '-1' in req_series_set:
                # -1 means 'all' - without any limit
                series_set = None
                self.series_limit = None
            else:
                # we process the set normally
                series_set = set()
                for series_id in req_series_set:
                    try:
                        series_id = int(series_id)
                    except:
                        pass
                    else:
                        if series_id in possible_series:
                            series_set.add(series_id)
                            continue
                    if series_dim:
                        # try to remap it
                        try:
                            texttoint = series_dim.texttointmap_set.get(
                                                                text=series_id)
                        except models.TextToIntMap.DoesNotExist:
                            pass
                        else:
                            series_set.add(texttoint.id)
        # time processing - display individual points, do a difference, etc.
        time_sampling =  None
        time_aggregate = None
        if not self.data_type.time_sampling_compatible():
            time_aggregate = Consts.TIME_AGGREG_NONE
            time_sampling = None
        else:
            if req_time_processing in DataQuery.TIME_AGGREG_VALUES:
                # all known integer values
                time_aggregate = req_time_processing
            else:
                # it is not a known integer value
                if req_time_processing not in self.VALID_TIME_PROCESSING_KEYS:
                    req_time_processing = self.DEFAULT_TIME_PROCESSING
            if time_aggregate is None:
                if req_time_processing == "diff":
                    time_aggregate = Consts.TIME_AGGREG_DIFF
                elif req_time_processing == "reldiff":
                    time_aggregate = Consts.TIME_AGGREG_DIFF_REL
                elif type(req_time_processing) in (str, unicode) and \
                     req_time_processing.startswith("i-"):
                    # parse individual definitions
                    sample_def = req_time_processing[2:]
                    if sample_def == "1d":
                        time_aggregate = Consts.TIME_AGGREG_NONE
                    else:
                        time_sampling = sample_def
                        if self.data_type.recurrent:
                            time_aggregate = Consts.TIME_AGGREG_SUM
                        else:
                            time_aggregate = Consts.TIME_AGGREG_LAST
                else:
                    logging.info("strange time processing requested '%s'",
                                 req_time_processing)
                    time_aggregate = Consts.TIME_AGGREG_NONE 
        # reference time
        available_dates = models.DataPoint.objects\
            .filter(data_type=self.data_type).aggregate(max_date=Max('date'),
                                                        min_date=Min('date'))
        end_date = None
        if req_ref_date:
            try:
                end_date = datetime.datetime.strptime(req_ref_date, "%Y-%m-%d")
            except Exception:
                logging.debug("Cannot parse date '%s'", req_ref_date)
                end_date = None
            else:
                end_date = end_date.date()
        if not end_date:
            end_date = available_dates['max_date']
            if end_date:
                end_date = end_date.date()
        # start time based on the selected range
        self.time_sampling = time_sampling
        rounding = self.get_time_rounding()
#        if req_date_range not in self.VALID_DATE_RANGE_KEYS:
#            req_date_range = self.DEFAULT_DATE_RANGE
        if not req_date_range:
            req_date_range = self.DEFAULT_DATE_RANGE
        start_date = DateManipulation.parse_time_range(end_date, req_date_range,
                                                       rounding=rounding)
        if rounding:
            end_date = DateManipulation.round_time(end_date, rounding, up=True)
                        
        logger.debug("selected date: %s - %s", start_date, end_date)
        # data scaling
        client_side_scaling = None
        if req_data_scaling in ("percent", Consts.SCALE_RELATIVE_100):
            data_scaling = Consts.SCALE_RELATIVE_100
            req_data_scaling = "percent"
        elif req_data_scaling in ("log",):
            data_scaling = Consts.SCALE_NORMAL
            client_side_scaling = "log"
        else:
            data_scaling = Consts.SCALE_NORMAL
            req_data_scaling = "normal"
        # summable_series
        summable_series = False
        if series_dim and isinstance(series_dim, models.Dimension):
            try:
                dtt = series_dim.dimensiontodatatype_set.get(
                                                    data_type=self.data_type)
            except models.DimensionToDataType.DoesNotExist:
                summable_series = False
            else:
                summable_series = dtt.partitioning
        # display as
        if req_display_as not in self.VALID_DISPLAY_AS_KEYS:
            logging.warn("Unsupported value for '%s' requested: %s",
                         self.KEY_DISPLAY_AS, req_display_as)
            display_as = self.DEFAULT_DISPLAY_AS
        else:
            display_as = req_display_as
        # store the results in attributes
        self.slicing_dim_values = slicing_dim_values
        self.slicing_dim = slicing_dim
        self.slicing_value = slicing_value
        self.start_date = start_date
        self.end_date = end_date
        self.time_aggregate = time_aggregate
        self.possible_series = possible_series
        self.series_set = series_set
        self.data_scaling = data_scaling
        self.summable_series = summable_series
        self.display_as = display_as
        self.client_side_scaling = client_side_scaling
        self.available_dates = available_dates
        # also store some input values
        self.req_date_range = req_date_range
        self.req_time_processing = req_time_processing
        self.req_data_scaling = req_data_scaling
        
        
    def get_time_rounding(self):
        if self.time_sampling and self.time_sampling[-1] in ("my"):
            return self.time_sampling[-1]
        return None
    
        
    def get_valid_time_ranges(self):
        max_date = self.available_dates['max_date']
        min_date = self.available_dates['min_date']
        if min_date and max_date:
            available_range = max_date - min_date
        else:
            available_range = datetime.timedelta(days=0)
        was_selected = False
        for value, name in self.DATE_RANGES:
            selected = False
            if value != "all":
                interval = DateInterval.from_string(value).get_approx_timedelta()
                if interval > available_range:
                    # this interval is not interesting
                    continue
            else:
                if not was_selected:
                    # this is 'all' and nothing else was selected
                    selected = True
            if available_range == datetime.timedelta(days=0) and value == "0d" \
               and self.req_date_range == self.DEFAULT_DATE_RANGE:
                # we use 0d as default when only one time record is present
                was_selected = True
                selected = True
            if value == self.req_date_range:
                was_selected = True
                selected = True
            yield (value, name, selected)

    
    def get_valid_time_processings(self):
        max_date = self.available_dates['max_date']
        min_date = self.available_dates['min_date']
        if min_date and max_date:
            available_days = (max_date - min_date).days + 1
        else:
            available_days = 0
        for value, name in self.TIME_PROCESSINGS:
            selected = False
            if value.startswith("i-"):
                days = DateInterval.from_string(value[2:]).get_approx_days()
                if days > available_days:
                    # this interval is not interesting
                    continue
            else:
                if available_days == 1:
                    # if only one day is available, we cannot do diff
                    continue
                elif self.data_type.recurrent:
                    # for recurrent data, there is not point in computing diff
                    continue
            if value == self.req_time_processing:
                selected = True
            yield (value, name, selected)
            
    
    def get_valid_display_as_options(self):
        return self.DISPLAY_AS_OPTS
    
    def get_valid_scaling_options(self):
        result = [("normal", ugettext("normal"))]
        log_ok = True
        rel_ok = (self.dim_count() > 0)
        if not self.time_sampling:
            if self.time_aggregate == Consts.TIME_AGGREG_DIFF:
                log_ok = False # there can be negative values in diff
            elif self.time_aggregate == Consts.TIME_AGGREG_DIFF_REL:
                log_ok = False # negative values again
                rel_ok = False # relative from relative data is too strange
        if not self.summable_series:
            rel_ok = False
        if rel_ok:
            result.append(("percent", ugettext("relative in %")))
        if log_ok:
            result.append(("log", ugettext("logarithmic")))
        return result
    
    def get_valid_slicing_values(self):
        series = self.slicing_dim_values
        revd = [(value.lower(), value, key) for key,value in series.iteritems()]
        revd.sort(key=lambda x: locale.strxfrm(x[0].encode('utf-8')))
        series = []
        for _ignore, key, value in revd:
            series.append((value, key))
        return series


    def get_possible_series(self):
        revd = [(value, key) for key,value in self.possible_series.iteritems()]
        revd.sort(key=lambda x: locale.strxfrm(x[0].encode('utf-8')))
        if self.series_set:
            zero_active = False
            series_set = self.series_set
        else:
            zero_active = True
            series_set = set()
        if len(revd) <= self.SERIES_LIMIT:
            series = [(0, ugettext("All"), zero_active)]
        else:
            series = [(-1, ugettext("All"), False),
                      (0, ugettext("Top %d") % self.SERIES_LIMIT, zero_active)]
        for key, value in revd:
            series.append((value, key, value in series_set))
        return series
        
    
    def create_data_query(self):
        # fixed values for 4D data slicing
        fixed_values = []
        for dim in self.dimensions:
            if dim == self.slicing_dim:
                fixed_values.append(self.slicing_value)
            else:
                fixed_values.append(None)
        args = (self.data_type, self.start_date, self.end_date)
        kwargs = dict(fixed_values=fixed_values,
                      time_aggregate=self.time_aggregate,
                      time_sampling=self.time_sampling)
        dqr = DataQuery(*args, **kwargs)
        return dqr
    
    def dim_count(self):
        return len(self.dimensions)
    
    def get_data(self):
        # preliminary checks
        if self.time_aggregate != Consts.TIME_AGGREG_NONE and \
           self.data_scaling != Consts.SCALE_NORMAL and \
           not self.time_sampling:
            # there are incompatible
            raise self.IncompatibleSettingsError()
        dquery = self.create_data_query()
        bundle = dquery.fetch_results()
        # we get is here before any sampling may mangle the data
        real_start_time, real_end_time = bundle.get_stored_time_range() 
        logger.debug(bundle)
        # do postprocessing
        postprocessing_attrs = {}
        if self.series_set:
            postprocessing_attrs['series_set'] = self.series_set
        elif self.display_as == "table_all":
            # no series limit here
            pass
        elif self.series_limit:
            # there could be different series limits for different types of 
            # data, but for now, we use one value for all
            if "_time" in bundle.dimensions:
                postprocessing_attrs['limit_series'] = self.series_limit
            else:
                if self.time_aggregate == Consts.TIME_AGGREG_NONE:
                    postprocessing_attrs['limit_series'] = self.series_limit
                else:
                    postprocessing_attrs['limit_series'] = self.series_limit
        postprocessing_attrs['scale'] = self.data_scaling
        postprocessing_attrs['create_other'] = self.include_other
        bundle.postprocess(**postprocessing_attrs)
        #pprint(bundle.data)
        # create dataTable
        series, series_desc = bundle.get_series_description()
        data_table = gviz_api.DataTable(series_desc)
        table_data = list(bundle.gen_data_table())
        data_table.LoadData(table_data)
        min_time = None
        max_time = None
        colors = None
        if bundle.key_dim == "_time": #series_desc["_main_key"][0] == "date":
            has_time_dim = True
            col_order, colors = self._sort_series_and_extract_color(series,
                                                                    series_desc)
            if table_data:
                min_time = real_start_time.date()
                max_time = real_end_time.date()
            order_data_by = "_main_key"
            dimension_x = series_desc["_main_key"][1]
            #self.dim_to_text(bundle.key_dim)
        else:
            has_time_dim = False
            col_order = sorted([key for key in series_desc if key != "_amount"],
                               reverse=True)
            col_order.append("_amount")
        if not min_time:
            if dquery.from_date:
                min_time = dquery.from_date
            else:
                min_time = self.start_date
        if not max_time:
            if dquery.to_date:
                max_time = dquery.to_date
            else:
                max_time = self.end_date
            order_data_by=None
            dimension_x = self.dim_to_text(bundle.key_dim)
        # amount dim modifier
        dim_y_mod = ""
        value_suffix = None
        if self.data_scaling == Consts.SCALE_RELATIVE_100:
            dim_y_mod = " - " + ugettext("Relative share [%]")
            value_suffix = "%"
        if self.time_aggregate == Consts.TIME_AGGREG_DIFF:
            dim_y_mod = " - "+ ugettext("Difference")
        elif self.time_aggregate == Consts.TIME_AGGREG_DIFF_REL:
            dim_y_mod = " - " + ugettext("Relative difference [%]")
            value_suffix = "%"
        result = self.ResultPackage()
        result.drq = self
        result.data_bundle = bundle
        result.data_table = data_table
        result.min_time = min_time
        result.max_time = max_time
        result.has_time_dim = has_time_dim
        result.column_order = col_order
        result.order_data_by = order_data_by
        result.dimension_x = dimension_x
        result.dimension_y = self.dim_to_text(bundle.amount_dim) + dim_y_mod
        result.value_suffix = value_suffix
        result.custom_colors = colors
        return result


    def _sort_series_and_extract_color(self, series, series_desc):
        if series:
            print series
            # lets see if we have some custom ordering or
            to_sort = []
            for name, color, order in series:
                to_sort.append((order, name, color))
            # here we sort in reverse order because hicharts does the display
            # in opposite order
            to_sort.sort(
                    key=lambda x: (-x[0],
                                   locale.strxfrm(x[1].lower().encode('utf-8')))
                    )
            sorted_series = []
            colors = []
            some_color = False
            for _ord, name, color in to_sort:
                sorted_series.append(name)
                colors.append(color)
                if color:
                    some_color = True
            if not some_color:
                colors = []
        else:
            sorted_series = sorted(
                    [key for key in series_desc if key != "_main_key"],
                    key=lambda x: locale.strxfrm(x.lower().encode('utf-8'))
                    )
            colors = []
        col_order = ["_main_key"] + sorted_series
        return col_order, colors
    

    def get_json_data(self):
        """
        Returns data suitable for json reply - does some more processing
        on the result of self.get_date
        """
        try:
            package = self.get_data()
        except DataRetreivalException:
            error = "Not enough data to display selected graph"
            ret = {"ok": False, "error": error}
            return ret
        except DataRequest.IncompatibleSettingsError:
            error = "Incompatible options selected"
            ret = {"ok": False, "error": error}
            return ret
        # serialize data
        json_data = package.data_table.ToJSon(columns_order=package.column_order,
                                              order_by=package.order_data_by)
        # chart title
        subvalue = self.slicing_value
        chart_title_parts = []
        if subvalue:
            try:
                subvalue_text = unicode(models.TextToIntMap.objects.get(id=subvalue))
            except models.TextToIntMap.DoesNotExist:
                subvalue_text = str(subvalue)
            chart_title_parts.append(subvalue_text)
        else:
            subvalue_text = None
        if package.min_time != package.max_time:
            t_rounding = self.get_time_rounding()
            part = _("from %(start_date)s to %(end_date)s") %\
                    {"start_date": DateManipulation.str_date(package.min_time,
                                                        rounding=t_rounding),
                    "end_date":  DateManipulation.str_date(package.max_time,
                                                        rounding=t_rounding)}
            if not chart_title_parts:
                part = part.capitalize()
            chart_title_parts.append(part)
        elif package.max_time:
            chart_title_parts.append(str(package.max_time))
        chart_title = ", ".join(chart_title_parts)
        # chart type
        chart_type = None
        if self.display_as in ("table","table_all"):
            chart_type = "table"
        else:
            if package.has_time_dim:
                if len(package.data_bundle.data) < 10:
                    chart_type = "column_chart"
                else:
                    chart_type = "line_chart"
            else:
                if self.time_aggregate == Consts.TIME_AGGREG_NONE:
                    chart_type = "column_chart"
                else:
                    chart_type = "bar_chart"
        # notes
        notes = []
        if self.time_aggregate in (Consts.TIME_AGGREG_DIFF,
                                   Consts.TIME_AGGREG_DIFF_REL):
            # in case of computing difference, it is important to warn that the
            # data does not fully match
            if self.end_date and self.end_date != package.max_time:
                notes.append(ugettext("Requested end time could not have "
                                      "been fully matched."))
            if self.start_date and self.start_date != package.min_time:
                notes.append(ugettext("Requested start time could not have "
                                      "been fully matched."))
        # problem description
        if package.data_table.NumberOfRows() == 0:
            if self.series_set:
                chart_text = ugettext("No data available for selected time "
                                      "and series.")
            else:
                chart_text = ugettext("No data available for selected time.")
        else:
            chart_text = None
        # serialize and return the data
        t_rounding = self.get_time_rounding()
        ret = {"ok": True,
               "data": json_data,
               "has_time": package.has_time_dim,
               "start_date": self.start_date,
               "end_date": self.end_date,
               "real_start_date": DateManipulation.str_date(package.min_time,
                                                            rounding=t_rounding),
               "real_end_date": DateManipulation.str_date(package.max_time,
                                                          rounding=t_rounding),
               "data_type": self.data_type.name,
               "subvalue": subvalue,
               "dimension_x": package.dimension_x,
               "dimension_y": package.dimension_y,
               "value_suffix": package.value_suffix,
               "summable_series": self.summable_series,
               "time_aggregate": self.time_aggregate,
               "chart_title": chart_title,
               "chart_type": chart_type,
               "notes": notes,
               "client_side_scaling": self.client_side_scaling,
               "chart_text": chart_text,
               "custom_colors": package.custom_colors,
               }
        return ret


    @classmethod
    def dim_to_text(cls, dim):
        if isinstance(dim, models.Dimension):
            return dim.name
        elif dim is None:
            return ""
        else:
            return unicode(cls.DIM_KEY_REMAP.get(dim, dim))

