# build-in modules
import os
import logging
import json
from optparse import make_option
import hashlib
import yaml

# Django imports
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.core.files import File

# local imports
from stats import models

# basic setup
logger = logging.getLogger("stats.update_db")

class Command(BaseCommand):
    
    args = "JSON or YAML file"
    
    option_list = BaseCommand.option_list + (
        make_option('-f', '--force-update',
            action='store_true',
            dest='force_update',
            default=False,
            help='Force update on instances different from db'),
        )
    
    def handle(self, *args, **options):
        if not args:
            raise CommandError('Give me a JSON or YAML file to load from')
        force_update = options.get('force_update', False)
        for arg in args:
            if not os.path.isfile(arg):
                raise CommandError('File "%s" does not exist' % arg)
            if arg.endswith(".json"):
                loader = json
            else:
                loader = yaml
            try:
                with file(arg, 'r') as infile:
                    data = loader.load(infile)
            except Exception as exc:
                raise CommandError('Cannot read file "%s" - error "%s"' % \
                                   (arg, exc))
            # dimensions go before data types
            statistics = data.get('statistics', {})
            for code_name, definition in statistics.iteritems():
                self._create_statistics(code_name, definition,
                                        force_update=force_update)
            # data types
            stat_groups = data.get("stat_groups", {})
            for code_name, definition in stat_groups.iteritems():
                self._create_stat_group(code_name, definition,
                                        force_update=force_update)

        
    @transaction.commit_on_success()
    def _create_statistics(self, code_name, definition, force_update=False):
        # default params
        default_params = definition.get("default_params")
        param_set = None
        if default_params:
            ps_code = default_params.get('code_name')
            if not ps_code:
                raise CommandError('code_name missing from "default_params"')
            try:
                param_set = models.ChartParamSet.objects.get(code_name=ps_code)
            except models.ChartParamSet.DoesNotExist:
                logger.debug("Creating ChartParamSet '%s'", ps_code)
                param_set = models.ChartParamSet(**default_params)
                param_set.save()
            else:
                logger.debug("ChartParamSet '%s' already exists", ps_code)
                if force_update:
                    updated = []
                    for name, value in default_params.iteritems():
                        old_value = getattr(param_set, name)
                        if old_value != value:
                            updated.append(name)
                            setattr(param_set, name, value)
                    if updated:
                        param_set.save()
                        logger.debug("Updated '%s' of param_set '%s'",
                                     ",".join(updated), ps_code)
        # the statistics itself
        attrs = dict(default_params=param_set)
        for lang, name in definition.get('name', {}).iteritems():
            attrs["name_"+lang] = name
        for lang, name in definition.get('info', {}).iteritems():
            attrs["info_"+lang] = name
        for lang, name in definition.get('methodology', {}).iteritems():
            attrs["methodology_"+lang] = name
        for lang, name in definition.get('stat_file', {}).iteritems():
            attrs["stat_file_"+lang] = File(file(name, 'r'))
        dt_code = definition.get("data_type")
        if dt_code is None:
            data_type = None
        else:
            try:
                data_type = models.DataType.objects.get(code_name=dt_code)
            except models.DataType.DoesNotExist:
                logger.error("DataType '%s' referenced from Statistics '%s' "
                             "does not exits", dt_code, code_name) 
                raise CommandError('Inconsistency in JSON data')
        attrs['data_type'] = data_type
        try:
            stat = models.Statistics.objects.get(code_name=code_name)
        except models.Statistics.DoesNotExist:
            logger.info("Creating Statistics '%s'", code_name)
            stat = models.Statistics(code_name=code_name, **attrs)
            stat.save()
        else:
            logger.debug("Statistics '%s' already exists", code_name)
            if force_update:
                updated = []
                for name, value in attrs.iteritems():
                    old_value = getattr(stat, name)
                    if hasattr(old_value, 'size'):
                        # this is a file?
                        if not self._compare_files(old_value, value):
                            updated.append(name)
                            setattr(stat, name, value)
                    else:
                        if old_value != value:
                            updated.append(name)
                            setattr(stat, name, value)
                if updated:
                    stat.save()
                    logger.debug("Updated '%s' of statistics '%s'",
                                 ",".join(updated), code_name)
        

    @transaction.commit_on_success()
    def _create_stat_group(self, code_name, definition, force_update=False):
        # the data type itself
        attrs = {}
        for lang, name in definition.get('name', {}).iteritems():
            attrs["name_"+lang] = name
        for lang, name in definition.get('info', {}).iteritems():
            attrs["info_"+lang] = name
        attrs['position'] = definition.get("position", 100)
        try:
            sgroup = models.StatGroup.objects.get(code_name=code_name)
        except models.StatGroup.DoesNotExist:
            logger.info("Creating StatGroup '%s'", code_name)
            sgroup = models.StatGroup(code_name=code_name, **attrs)
            sgroup.save()
        else:
            logger.debug("StatGroup '%s' already exists", code_name)
            if force_update:
                updated = []
                for name, value in attrs.iteritems():
                    if getattr(sgroup, name) != value:
                        updated.append(name)
                        setattr(sgroup, name, value)
                if updated:
                    sgroup.save()
                    logger.debug("Updated '%s' of StatGroup '%s'",
                                 ",".join(updated), code_name)
        # connect it with dimensions
        for pos, stat_code in enumerate(definition.get('statistics', [])):
            try:
                stat = models.Statistics.objects.get(code_name=stat_code)
            except models.Statistics.DoesNotExist:
                logger.error("Statistics '%s' referenced in StatGroup '%s' does "
                             "not exist", stat_code, code_name)
                raise CommandError("Error in input file.")
            else:
                # connect Dimensions with Datatype
                try:
                    sgts = models.StatGroupToStatistics.objects.get(
                                            stat_group=sgroup, statistics=stat)
                except models.StatGroupToStatistics.DoesNotExist:
                    logger.info("Connecting StatGroup '%s' with Statistics '%s'",
                                code_name, stat_code)
                    sgts = models.StatGroupToStatistics(stat_group=sgroup,
                                                        statistics=stat,
                                                        position=pos)
                    sgts.save()
                else:
                    logger.debug("StatGroup '%s' already connected to Statistics "
                                 "'%s'", code_name, stat_code)
                    if sgts.position != pos:
                        logger.debug("Updating position of '%s'", stat_code)
                        sgts.position = pos
                        sgts.save()
                        
    def _compare_files(self, f1, f2):
        def do_hash(fobj):
            hasher = hashlib.sha1()
            hasher.update(fobj.read())
            return hasher.hexdigest()
        return do_hash(f1) == do_hash(f2)