# django imports
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.conf import settings

# third party imports
from transmeta import TransMeta

# local imports
from common import Consts


class DashboardChart(models.Model):
    """
    Represents a chart to be shown at the dashboard
    """
    __metaclass__ = TransMeta

    code_name = models.SlugField(unique=True)
    name = models.CharField(max_length=64)
    statistics = models.ForeignKey("Statistics")
    position = models.IntegerField(default=0)
    params = models.OneToOneField("ChartParamSet")

    class Meta:
        verbose_name = _("Dashboard chart")
        verbose_name_plural = _("Dashboard chart")
        translate = ('name',)


class ChartParamSetManager(models.Manager):
    def get_by_natural_key(self, code_name):
        return self.get(code_name=code_name)

class ChartParamSet(models.Model):
    """
    Represents a reference to specific params for a chart
    """
    __metaclass__ = TransMeta
    objects = ChartParamSetManager()
        
    TIME_PROC_CHOICES = (
        (Consts.TIME_AGGREG_NONE, _("Normal")),
        (Consts.TIME_AGGREG_DIFF, _("Difference")),
        (Consts.TIME_AGGREG_DIFF_REL, _("Relative difference")),
        )
    
    TIME_AGGREG_CHOICES = (
        (),
        )
    
    DATA_SCALING_CHOICES = (
        (Consts.SCALE_NORMAL, _("Normal")),
        (Consts.SCALE_RELATIVE_100, _("Relative in %")),
        )
    
    CLIENT_DATA_SCALING_CHOICES = (
        (Consts.CLIENT_SCALE_NORMAL, _("Normal")),
        (Consts.CLIENT_SCALE_LOG, _("Logarithmic")),
        )
        
    code_name = models.SlugField(unique=True)
    ref_date = models.CharField(max_length=16, default="")
    date_range = models.CharField(max_length=16)
    time_processing = models.IntegerField(choices=TIME_PROC_CHOICES,
                                          default=Consts.TIME_AGGREG_NONE)
    time_sampling = models.CharField(max_length=16, default="")
    data_scaling = models.IntegerField(choices=DATA_SCALING_CHOICES,
                                       default=Consts.SCALE_NORMAL)
    client_data_scaling = models.IntegerField(choices=CLIENT_DATA_SCALING_CHOICES,
                                              default=Consts.CLIENT_SCALE_NORMAL)
    include_other = models.BooleanField(default=False,
                        help_text=_("Should 'other' be created when limiting "
                                    "the number of series"))
    series_limit = models.IntegerField(default=0)
    one_series = models.CharField(max_length=255, null=True)
    slicing_dim = models.IntegerField(null=True)
    slicing_value = models.CharField(max_length=255, null=True)
    
    def natural_key(self):
        return self.code_name
        
    def get_data_request_params(self):
        # we import here to avoid circular import because data_request
        # imports models
        from data_request import DataRequest
        if self.time_processing == Consts.TIME_AGGREG_NONE:
            if self.time_sampling:
                time_processing = "i-" + self.time_sampling
            else:
                time_processing = DataRequest.DEFAULT_TIME_PROCESSING
        else:
            time_processing = self.time_processing
        # scaling is synthetised from two values
        if self.data_scaling == Consts.SCALE_NORMAL and \
           self.client_data_scaling == Consts.CLIENT_SCALE_LOG:
            data_scaling = "log"
        else:
            data_scaling = self.data_scaling
        params = {DataRequest.KEY_REF_DATE: self.ref_date,
                  DataRequest.KEY_DATE_RANGE: self.date_range,
                  DataRequest.KEY_TIME_PROCESSING: time_processing,
                  DataRequest.KEY_DATA_SCALING: data_scaling,
                  DataRequest.KEY_SERIES_SET: self.one_series,
                  DataRequest.KEY_SERIES_LIMIT: self.series_limit,
                  DataRequest.KEY_INCLUDE_OTHER: self.include_other,
                  DataRequest.KEY_SLICING_DIM: self.slicing_dim,
                  DataRequest.KEY_SLICING_VALUE: self.slicing_value,
                  }
        return params
    


class StatGroup(models.Model):
    """
    Represents a group of statistics
    """
    __metaclass__ = TransMeta
        
    code_name = models.SlugField(unique=True)
    name = models.CharField(max_length=64)
    statistics = models.ManyToManyField("Statistics",
                                        through="StatGroupToStatistics")
    position = models.PositiveIntegerField()
    info = models.TextField()
    
    class Meta:
        verbose_name = _("Statistics group")
        verbose_name_plural = _("Statistics group")
        translate = ('name', 'info')
        ordering = ("position",)
        
    def get_statistics(self):
        return self.statistics.order_by('statgrouptostatistics__position')
    
    def get_non_empty_statistics(self):
        ret = []
        for stat in self.statistics.order_by('statgrouptostatistics__position')\
                                                .select_related("data_type"):
            # unfortunately this does a query for each stat.data_type to
            # determine if it has data;
            # I tried to use annotations to add the count of datapoints into
            # the query, but it resulted in a huge and very slow SQL; thus
            # this naive approach is used and we rely on caching 
            if stat.data_type is None or stat.data_type.has_data():
                ret.append(stat)
        return ret
                
        


class StatGroupToStatistics(models.Model):
    """
    Mapping between StatGroup and Statistics providing an extra position
    attribute
    """
    stat_group = models.ForeignKey(StatGroup)
    statistics = models.ForeignKey('Statistics')
    position = models.PositiveSmallIntegerField()
    
    class Meta:
        unique_together = (("stat_group", "statistics", "position"),)
        verbose_name = _("StatGroup to Statistics mapping")
        verbose_name_plural = _("StatGroup to Statistics mappings")
        ordering = ("position",)



class StatisticsManager(models.Manager):
    def get_by_natural_key(self, code_name):
        return self.get(code_name=code_name)

class Statistics(models.Model):
    """Represents one statistics such as it is displayed on the web"""
    __metaclass__ = TransMeta
    objects = StatisticsManager()
    
    code_name = models.SlugField(unique=True)
    name = models.CharField(verbose_name=_("Name"), max_length=64)
    data_type = models.ForeignKey("DataType", verbose_name=_("Data type"),
                                  null=True, blank=True)
    stat_file = models.FileField(verbose_name=_("File"),
                                 upload_to="offline_stats",
                                 null=True, blank=True)
    info = models.TextField(verbose_name=_("Info"))
    methodology = models.TextField(verbose_name=_("Methodology"))
    default_params = models.ForeignKey(ChartParamSet, null=True, blank=True)
    
    class Meta:
        verbose_name = _("Statistics")
        verbose_name_plural = _("Statistics")
        translate = ('name', 'info', 'methodology', 'stat_file')
        
    def natural_key(self):
        return self.code_name
    
    def __unicode__(self):
        if self.name:
            return self.name
        return self.code_name
    
    def is_offline(self):
        if self.data_type is None:
            return True
        return False
    
    def has_localized_file(self):
        if self.stat_file:
            name = None
            for lang_code, _lang_name in settings.LANGUAGES:
                loc_file = getattr(self, "stat_file_"+lang_code)
                if name and loc_file.name != name:
                    return True
                name = loc_file.name
            return False
        return None
    
    def get_localized_files(self):
        files = []
        for lang_code, _lang_name in settings.LANGUAGES:
            loc_file = getattr(self, "stat_file_"+lang_code)
            files.append((lang_code, loc_file))
        return files



class DataType(models.Model):
    """Represents one type of data, that is a relationship between dimensions"""
    __metaclass__ = TransMeta
    
    code_name = models.SlugField(unique=True)
    name = models.CharField(max_length=64)
    amount_dim = models.ForeignKey('Dimension',
                                   related_name="amounttodatatype_set")
    recurrent = models.BooleanField(default=False,
                                    help_text="Is the amount counted from 0 in "
                                              "each time period?")
    summable_periods = models.BooleanField(default=True,
                        help_text="When recurrent, is it possible to sum "
                                  "individual values to get value for longer "
                                  "period?")
    dimensions = models.ManyToManyField("Dimension",
                                        through="DimensionToDataType")

    class Meta:
        verbose_name = _("Data type")
        verbose_name_plural = _("Data types")
        translate = ('name',)
        ordering = ("code_name",)
        
    def __unicode__(self):
        return self.name
        
    def get_dimensions(self):
        result = [d2dt.dimension
                  for d2dt in self.dimensiontodatatype_set.all()\
                            .select_related("dimension").order_by("position")]
        return result
    
    def time_sampling_compatible(self):
        if self.recurrent and not self.summable_periods:
            return False
        return True
    
    def has_data(self):
        return self.datapoint_set.all().exists()

#
class Dimension(models.Model):
    """One dimension of a certain type"""
    __metaclass__ = TransMeta
    
    TYPE_INT = 1
    TYPE_TEXT = 2
    
    DIMENSION_TYPE_CHOICES = ((TYPE_INT, _('integer')),
                              (TYPE_TEXT, _('text')),
                              )
    
    code_name = models.SlugField(unique=True)
    name = models.CharField(max_length=64)
    type = models.PositiveSmallIntegerField(choices=DIMENSION_TYPE_CHOICES)
    
    class Meta:
        verbose_name = _("Dimension")
        verbose_name_plural = _("Dimensions")
        translate = ('name',)
        ordering = ("code_name",)
        
    def __unicode__(self):
        return u"{0} ({1})".format(self.name, self.code_name)

    @classmethod
    def get_type_by_name(cls, name):
        if name in ("int", "integer"):
            return cls.TYPE_INT
        elif name in ("str", "string", "text"):
            return cls.TYPE_TEXT
        raise ValueError("Unsupported dimension type '%s'" % name)
    
    def get_type_string(self):
        if self.type == self.TYPE_INT:
            return "integer"
        elif self.type == self.TYPE_TEXT:
            return "string"
        raise ValueError("Unsupported dimension type '%d'" % self.type)
    
    def get_str_to_int_remap(self):
        result = dict([tuple(values)
                       for values in 
                       self.texttointmap_set.all().values_list("text","id")])
        return result
    
    def get_int_to_str_remap(self, lang="en"):
        result = {}
        for pk, text, local in self.texttointmap_set.all().\
                                values_list("id","text","text_local_"+lang):
            string = local if local else text
            result[pk] = string
        return result
    
    def get_int_to_str_and_custom_values_remap(self, lang="en"):
        result = {}
        for pk, text, local, cclr, cord in self.texttointmap_set.all().\
                                values_list("id","text","text_local_"+lang,
                                            "custom_color","custom_order"):
            string = local if local else text
            result[pk] = (string, cclr, cord)
        return result

    def get_int_to_locale_remap(self):
        result = {}
        for values in self.texttointmap_set.all().values():
            new_values = {}
            for key, val in values.iteritems():
                if key == "text":
                    new_values[None] = val
                elif key.startswith("text_local_"):
                    new_values[key[11:]] = val
            result[values['id']] = new_values
        return result

    def remap_value(self, value):
        if self.type == self.TYPE_INT:
            return value
        ttim = self.texttointmap_set.get(id=value)
        return ttim
    
    def value_to_int(self, value):
        if self.type == self.TYPE_INT:
            return value
        try:
            ttim = self.texttointmap_set.get(text=value)
            return ttim.id
        except TextToIntMap.DoesNotExist:
            return None


class DimensionToDataType(models.Model):
    """
    Mapping between data type and its dimensions providing an extra position
    attribute
    """
    data_type = models.ForeignKey(DataType)
    dimension = models.ForeignKey(Dimension)
    position = models.PositiveSmallIntegerField()
    partitioning = models.BooleanField(
                        default=True, 
                        help_text=_("Does this dimension "
                                    "partition the data into summable parts"))
    
    class Meta:
        unique_together = (("data_type", "dimension", "position"),)
        verbose_name = _("Dimension to data type mapping")
        verbose_name_plural = _("Dimension to data type mappings")
        ordering = ("position",)



class DataPoint(models.Model):
    """One data point"""

    data_type = models.ForeignKey(DataType)
    date = models.DateTimeField()
    amount = models.BigIntegerField()
    value1 = models.IntegerField(null=True)
    value2 = models.IntegerField(null=True)
    value3 = models.IntegerField(null=True)
    
    class Meta:
        unique_together = (("data_type", "date", "value1", "value2", "value3"),)
        verbose_name = _("Data point")
        verbose_name_plural = _("Data points")


class DataPointAggreg(models.Model):
    """One data point aggregated in time from several data points;
    serves mostly for performance improvement;
    
    time_period - a text value describing the aggregated interval;
                  must be compatible with time_sampling used above
    amount - stores the value obtained by the preferred aggregation method for
             this data type
    amount_x - stores amount values aggregated using other methods, if they make
               sense, otherwise it is null
    """

    data_type = models.ForeignKey(DataType)
    date = models.DateTimeField()
    time_period = models.CharField(max_length=16)
    amount = models.BigIntegerField()
    value1 = models.IntegerField(null=True)
    value2 = models.IntegerField(null=True)
    value3 = models.IntegerField(null=True)
    # optional values for different aggregation methods
    amount_min = models.BigIntegerField(null=True)
    amount_max = models.BigIntegerField(null=True)
    amount_avg = models.BigIntegerField(null=True)
    amount_first = models.BigIntegerField(null=True)
    amount_last = models.BigIntegerField(null=True)
    amount_sum = models.BigIntegerField(null=True)
    amount_median = models.BigIntegerField(null=True)
    amount_middle = models.BigIntegerField(null=True)
    
    class Meta:
        unique_together = (("data_type", "date", "time_period", "value1",
                            "value2", "value3"),)
        verbose_name = _("Data point aggregation")
        verbose_name_plural = _("Data points aggregations")



class TextToIntMap(models.Model):
    """
    Mapping between string and integer values for a specific dimension
    """
    __metaclass__ = TransMeta
        
    id = models.AutoField(primary_key=True)
    dimension = models.ForeignKey(Dimension)
    text = models.CharField(max_length=255, db_index=True)
    text_local = models.CharField(max_length=255)
    custom_color = models.CharField(max_length=16, default="")
    custom_order = models.IntegerField(default=0)

    class Meta:
        verbose_name = _("Text to int map")
        verbose_name_plural = _("Text to int maps")
        unique_together = (("dimension", "text"),)
        translate = ("text_local",)
        
    def __unicode__(self):
        if self.text_local:
            return self.text_local
        return self.text
