# encoding: utf-8

# built-in modules
import datetime
import json
import logging
import copy

# Django imports
from django.shortcuts import render
from django.http import HttpResponse, Http404
from django.core.cache import cache
from django.utils.translation import get_language
from django.conf import settings 

# local imports
from stats import models
from data_mining import DataRetreivalException
from data_request import DataRequest
from common import create_cache_key, querydict_to_dict

logger = logging.getLogger("stats.views")


def dashboard(request):
    chart_refs = models.DashboardChart.objects.all().order_by("position")
    data = dict(chart_refs=chart_refs)
    return render(request, 'dashboard.html', data)

def all_statistics(request):
    stat_groups = _get_all_stat_groups()
    data = dict(stat_groups=stat_groups)
    return render(request, 'all_statistics.html', data)



def statistics_page(request, stat_code_name):
    try:
        stat = models.Statistics.objects.select_related('default_params')\
                                                .get(code_name=stat_code_name)
    except models.Statistics.DoesNotExist:
        raise Http404
    # let's see if we have an online or offline statistics
    if stat.data_type is None:
        if stat.stat_file is None:
            # there must be a file if there is no data_type, otherwise this
            # stat is not active
            raise Http404
        stat_groups = _get_all_stat_groups()
        data = dict(statistics=stat, stat_groups=stat_groups)
        return render(request, "offline_statistics_page.html", data)
    else:
        drq = _get_data_request(request.GET, stat.code_name)
        stat_groups = _get_all_stat_groups()
        data = dict(data_request=drq, stat_groups=stat_groups)
        data.update(drq.available_dates)
        return render(request, "statistics_page.html", data)


def animation_test(request):
    return render(request, "animation_test.html")

def error500(request):
    return render(request, "500.html")

## ---------- JSON FOLLOWS ----------------

def json_data(request, stat_code_name):
    cache_key = create_cache_key("json_data", stat_code_name, request.GET)
    ret = cache.get(cache_key)
    if not ret:
        logger.debug("cache miss for %s", cache_key)
        drq = _get_data_request(request.GET, stat_code_name)
        ret = drq.get_json_data()
        cache.set(cache_key, ret)
    else:
        logger.debug("cache hit for %s", cache_key)
    return HttpResponse(json.dumps(ret, default=_json_encode),
                        mimetype="application/javascript")
    
#@cache_page(60*10)
def dashchart_data(request, ref_code_name):
    lang = get_language()
    cache_key = create_cache_key("json_data_dash", ref_code_name, lang)
    ret = cache.get(cache_key)
    if not ret:
        logger.debug("cache miss for %s", cache_key)
        try:
            dash_chart = models.DashboardChart.objects.select_related().\
                                                    get(code_name=ref_code_name)
        except models.DashboardChart.DoesNotExist:
            raise Http404
        params = dash_chart.params.get_data_request_params()
        drq = _get_data_request(params, dash_chart.statistics)
        ret = drq.get_json_data()
        ret['chart_title'] = dash_chart.name
        cache.set(cache_key, ret)
    else:
        logger.debug("cache hit for %s", cache_key)
    response = HttpResponse(json.dumps(ret, default=_json_encode),
                            mimetype="application/javascript")
    if hasattr(settings, "ACCESS_CONTROL_ALLOW_ORIGIN") and \
       settings.ACCESS_CONTROL_ALLOW_ORIGIN:
        response['Access-Control-Allow-Origin'] = \
                                          settings.ACCESS_CONTROL_ALLOW_ORIGIN
    return response 

def csv_data(request, stat_code_name):
    drq = _get_data_request(request.GET, stat_code_name)
    # fetch the data from the db
    try:
        package = drq.get_data()
    except DataRetreivalException:
        raise Http404
    return_data = package.data_table.ToCsv(separator="|",
                                           columns_order=package.column_order,
                                           order_by=package.order_data_by)
    res = HttpResponse(return_data, mimetype="text/csv")
    res["Content-Disposition"] = "attachment; filename=%s.csv" % stat_code_name
    return res


def json_slices(request, stat_code_name):
    drq = _get_data_request(request.GET, stat_code_name)
    series = drq.get_valid_slicing_values()
    ret = {"ok": True, "data": series, "slicing_value": drq.slicing_value}
    return HttpResponse(json.dumps(ret), mimetype="application/javascript")



def json_series(request, stat_code_name):
    drq = _get_data_request(request.GET, stat_code_name)
    series = []
    for key, value, sel in drq.get_possible_series():
        series.append((key, value, sel))
    data = dict(series=series, summable=drq.summable_series,
                scales=drq.get_valid_scaling_options())
    ret = {"ok": True, "data": data}
    return HttpResponse(json.dumps(ret), mimetype="application/javascript")

# HELPER FUNCTIONS

def _json_encode(value):
    if isinstance(value, datetime.date):
        return unicode(value)
    elif isinstance(value, datetime.datetime):
        return unicode(value)
    raise TypeError("Cannot serialize %s" % value)

# caching stuff

def _get_all_stat_groups():
    cache_key = create_cache_key("main", "stat_groups")
    data = cache.get(cache_key)
    if not data:
        data = []
        for stat_group in models.StatGroup.objects.all():
            data.append(dict(group=stat_group,
                             stats=stat_group.get_non_empty_statistics()))
        cache.set(cache_key, data)
    return data
        
        
def _get_data_request(params, stat_code_name):
    params = querydict_to_dict(params)
    if "lang" not in params:
        params["lang"] = get_language()
    cache_key = create_cache_key("data", stat_code_name, params)
    logger.debug("cache search for %s", cache_key)
    drq = cache.get(cache_key)
    if not drq:
        logger.debug("cache miss for %s", cache_key)
        drq = DataRequest(params, stat_code_name)    
        cache.set(cache_key, drq)
    else:
        logger.debug("cache hit for %s", cache_key)
    return drq

def crash(request):
    return XXX